{{=========================================================================== 

** Texas Instruments SN76489 Chip Sound Driver **

AUTHOR: Francesco de Simone, www.digimorf.com
LAST MODIFIED: 06.13.2011
VERSION 1.9

FILENAME: SN76489_019.spin

COMMENTS:
This chip has been used in arcade machines, home computers and game consoles of the eighties.
This driver generates sound through the audio pin, and emulates the chip Texa Instruments SN76489.
The PROPC3 uses the same design for audio as do most of the Propeller development boards.
Its simply a RC integrator passed by a capacitor to the audio port.

This is the first "educational" version with only one tone channel:

- 1 tone generator that produce square wave audio, with a frequency ranging from
109Hz to teorically 111860 Hz, but pratically it's only possible with an output sample rate higher than 44100Hz

- 1 volume attenuator which ranges from 0 to 15, where 0 is no attenuation and 15 is full attenuation.  

REQUIRES: Parallax PROPC3 Board

LICENSE: See end of file for terms of use

FIXES: Noise channel rendering

BUGS: It's possible to hear differences between real hardware sound and this emulation.
Timings have to be fixed and the main flow can be reduced. Actually, at this stage, this
driver needs to be tested with other emulated chips.

=========================================================================== }}

CON
  _clkmode = xtal1 + pll16x ' enable external clock and pll times 16
  _xinfreq = 5_000_000      ' set frequency to 5 MHZ

  SAMPLE_RATE = 44_100 
  COG_CLOCK = 80_000_000
  
VAR
  long  cogon, cog   ' Pointer for cog
  long  Per_Div0, Vol_Att_0, Per_Div1, Vol_Att_1, Per_Div2, Vol_Att_2 ' temp values for emulation
  long  Per_Div3, Vol_Att_3, noiseTap, noiseLFSR ' temp values for emulation   
  long  Latch_Chan, Latch_Volume, Chan_Tmp, Div_Tmp ' temp values for emulation   
  
  long  Registers[4] ' Memory mapped registers for the chip

OBJ   
  'UART     : "FullDuplexSerial" 
  
PUB start ( sound_ptr ) : okay
{{
Starts the chip sound driver on a new cog.

    sound_ptr:      The I/O pin to send audio to ( always 24 on the C3 )
    returns:        false if no cog available
}}
  stop  ' Stops eventual actual running cog
  Audio_pin := sound_ptr ' use the audio pin
  
  okay := cogon := ( cog := cognew( @entry, @Registers[0] )) > 0
  reset
  {
  UART.start(31,30,0,115200)      ' start the uart Tx and Rx and set the baud rate
  UART.tx($0D)
  UART.tx($0A)
  UART.str(string("SN76849 emulation"))
  UART.tx($0D)
  UART.tx($0A)
  UART.str(string("stream received:")) 
  UART.tx($0D)
  UART.tx($0A)
  }
PUB stop
{{
Stops the chip sound driver. Frees a cog.
}}

  if cogon~
    cogstop(cog)

PUB write ( data )
{{
  Write a data byte to the chip   
  Data byte can be:
   
  %1cctdddd
    |||````-- Data
    ||`------ Type
    ``------- Channel

  %0-DDDDDD
    |``````-- Data
    `-------- Unused    
}}
  'UART.hex( data,2 )
  'UART.tx ( 32 )
  
  if( data &%1000_0000 == %1000_0000 )  ' Check if byte is a register byte or data byte
    Chan_Tmp := ( data >> 5 ) & $3                         

    case Chan_Tmp
        0: Div_Tmp := ( Per_Div0 & $fff0 )  ' Update lower 4 bits of the period
           Div_Tmp := Div_Tmp | ( data &$F )
           
        1: Div_Tmp := ( Per_Div1 & $fff0 )
           Div_Tmp := Div_Tmp | ( data &$F )
           
        2: Div_Tmp := ( Per_Div2 & $fff0 )
           Div_Tmp := Div_Tmp | ( data &$F )
            
        3: Div_Tmp := ( Per_Div3 & $fff0 )
           Div_Tmp := Div_Tmp | ( data &$F )                                          

    Latch_Chan := Chan_Tmp     ' Latch channel number for future accesses   
    if ( data & $10 )          ' Latch volume for future accesses
        Latch_Volume := true
    else
        Latch_Volume := false
          
  else 
    Chan_Tmp := Latch_Chan
    case Chan_Tmp 
        0: Div_Tmp := ( Per_Div0 & $F )
           Div_Tmp := Div_Tmp | (( data & $3F ) << 4) ' Update higher 6 bits of the period
           
        1: Div_Tmp := ( Per_Div1 & $F )
           Div_Tmp := Div_Tmp | (( data & $3F ) << 4)
           
        2: Div_Tmp := ( Per_Div2 & $F )
           Div_Tmp := Div_Tmp | (( data & $3F ) << 4)
           
        3: Div_Tmp := ( Per_Div3 & $F )
           Div_Tmp := Div_Tmp | (( data & $3F ) << 4)
           
  if ( Latch_Volume )              ' Store volume attenuators
    case Chan_Tmp
        0: Vol_Att_0 := data & $F                                                                                                
        1: Vol_Att_1 := data & $F                                                                                               
        2: Vol_Att_2 := data & $F
        3: Vol_Att_3 := data & $F                                                                                                    
  else
    case Chan_Tmp                  ' Store periods
        0: Per_Div0 := Div_Tmp                                                                                                 
        1: Per_Div1 := Div_Tmp                                                                                               
        2: Per_Div2 := Div_Tmp                                                                                                
        3: Per_Div3 := Div_Tmp                                           

    if ( Chan_Tmp == 3 )           ' Compute Noise Channel
        if (( Div_Tmp >> 2 ) & 1 == 1 )
          noiseTap := 9
        else
          noiseTap := 1
          
        noiseLFSR := $8000

  ' Use only two longs to store each parameters to save space and memory access time
  Registers[0] := ( Vol_Att_1 << 28 ) | ( Per_Div1 << 16 ) | ( Vol_Att_0 << 12 ) | Per_Div0 ' refresh memory mapped registers
  Registers[1] := ( noiseTap << 24 ) | ( Vol_Att_3 << 20 ) | ( Per_Div3 << 16 ) | ( Vol_Att_2 << 12 ) | Per_Div2 ' refresh memory mapped registers
  Registers[2] := noiseLFSR
  
PUB reset
  Registers[0] := ( $F << 28 ) | ( $0 << 16 ) | ( $F << 12 ) | $0 ' refresh memory mapped registers 
  Registers[1] := ( $9 << 24 ) | ( $F << 20 ) | ( $3FF << 16 ) | ( $F << 12 ) | $3FF ' refresh memory mapped registers
  Registers[2] := $8000
  
DAT org 
entry
                     mov    cnt_accu, cnt                ' init counter                                                                                                                                           
                     add    cnt_accu, cnt_for_sample
                     
                     mov    t1, #1                       ' convert pin # into bit position
                     shl    t1, Audio_pin                ' audio pin taken from the caller
                                                                                                                                                                                      
                     or     dira, t1                     ' set general direction of I/O to output                                                                                                                
                     movs   ctra, Audio_pin              ' set counter A pin to the audio pin ( PROPC3 = 24 )                                                                                                        
                     movi   ctra, #%0_00110_000          ' duty cycle mode "00110", single ended

refresh              mov    chip_system, par             ' get caller parameters pointer into a variable
' CH 1 --------------------------------------------------------------------------------------------------------
                      ' refresh period register                                                              
                     rdlong t1, chip_system              ' read registers from the shared memory              
                     mov    psg_period_1, t1             ' put it into period                                 
                     and    psg_period_1, Period_mask    ' mask to 10 bits                                    
                     shl    psg_period_1, cnt_precision  ' gain some more space for counting reasons          
                     ' refresh attenuation register                                                          
                     shr    t1, #12                      ' align 12 bits to the right                          
                     mov    psg_att_1, t1                ' put it in the attenuator register                  
                     and    psg_att_1, #$F               ' mask to 4 bits                                      
' CH 2 --------------------------------------------------------------------------------------------------------
                      ' refresh period register
                     shr    t1, #4                       ' align 4 bits to the right          
                     mov    psg_period_2, t1             ' put it into period 
                     and    psg_period_2, Period_mask    ' mask to 10 bits
                     shl    psg_period_2, cnt_precision  ' gain some more space for counting reasons 
                     ' refresh attenuation register
                     shr    t1, #12                      ' align 12 bits to the right 
                     mov    psg_att_2, t1                ' Put it in the attenuator register
                     and    psg_att_2, #$F               ' mask to 4 bits                   
' CH 3 --------------------------------------------------------------------------------------------------------
                     add    chip_system, #4              ' put ahead the pointer to the shared memory
                     rdlong t1, chip_system              ' read registers from the shared memory
                     ' refresh period register  
                     mov    psg_period_3, t1             ' put it into period
                     and    psg_period_3, Period_mask    ' mask to 10 bits
                     shl    psg_period_3, cnt_precision  ' gain some more space for counting reasons  
                     ' refresh attenuation register
                     shr    t1, #12                      ' align 12 bits to the right        
                     mov    psg_att_3, t1                ' put it in the attenuator register
                     and    psg_att_3, #$F               ' mask to 4 bits                                      
' CH 4 --------------------------------------------------------------------------------------------------------
                     shr    t1, #4                       ' align 4 bits to the right    
                     mov    psg_period_4, t1             ' put it into noise register
                     and    psg_period_4, #$3            ' mask to 4 bits                        
                     ' refresh attenuation register
                     shr    t1, #4                       ' align 4 bits to the right    
                     mov    psg_att_4, t1                ' put it in the attenuator register
                     and    psg_att_4, #$F               ' mask to 4 bits                        
                     ' refresh tap bits
                     shr    t1, #4                       ' align 4 bits to the right
                     mov    TapBits, t1                  ' put it in the tap bits register
                     and    TapBits, #$F                 ' mask to 4 bits
                     ' refresh LFSR
                     add    chip_system, #4              ' put ahead the pointer to the shared memory
                     rdlong t1, chip_system              ' read registers from the shared memory
                     mov    ShiftReg, t1                 ' put into LFSR
                     ' compute noise
                     cmp    psg_period_4, #3       wc,wz ' sheck if the frequency is internal or external driven
         if_nc_and_z mov    psg_period_4, psg_period_3   ' if external ( 3 ) use frequency period of channel 3
         if_c_and_nz mov    per_tmp, #$10                ' if internal ( <= 2 ) calculate the right frquency period
         if_c_and_nz shl    per_tmp, psg_period_4        ' 16, 32, 64
         if_c_and_nz mov    psg_period_4, per_tmp        ' use period
         if_c_and_nz shl    psg_period_4, cnt_precision
' -------------------------------------------------------------------------------------------------------------
tone1                ' GET ATTENUATION VALUE
                     mov    tab_ptr, #att_table          ' create a pointer to the attenuation table                                    
                     add    tab_ptr, psg_att_1           ' add the correct index to that pointer                                        
                     movs   :read1, tab_ptr              ' move the pointer to the source field of instruction at :read position        
:read1               mov    att_tmp_1, 0                 ' moved the value in the source field ( 0 is a place holder ) to tab_data      
                    ' SWITCH OSCILLATOR                                                                                                 
                     cmp    psg_period_1, #1          wc ' check if period is 0, if so turn off oscillator                              
                if_c mov    psg_out_1, att_tmp_1         ' and use only attenuator                                                      
                if_c jmp    #tone2                       ' jump to the next step                                                        
                    ' CALCULATE OSCILLATOR                                                                                              
                     sub    psg_cnt_1, psg_SubUnit    wc ' decrease counter andwrite c if < 0                                           
                if_c mov    psg_cnt_1, psg_period_1      ' else resets counter to the period value                                      
                if_c xor    psg_osc_1, #1             wz ' z = 0 only when osc = 0 because of 0 xor 1 = 1, and 1 xor 1 = 0              
                if_c negz   psg_out_1, att_tmp_1         ' so we negate the output when z is not 0, ( when osc = 0 ) }                                                                  
' -------------------------------------------------------------------------------------------------------------
tone2                ' GET ATTENUATION VALUE             
                     mov    tab_ptr, #att_table          ' same as previous               
                     add    tab_ptr, psg_att_2           
                     movs   :read2, tab_ptr              
:read2               mov    att_tmp_2, 0                 
                    ' SWITCH OSCILLATOR                  
                     cmp    psg_period_2, #1          wc 
                if_c mov    psg_out_2, att_tmp_2         
                if_c jmp    #tone3                       
                    ' CALCULATE OSCILLATOR               
                     sub    psg_cnt_2, psg_SubUnit    wc 
                if_c mov    psg_cnt_2, psg_period_2      
                if_c xor    psg_osc_2, #1             wz 
                if_c negz   psg_out_2, att_tmp_2         
' -------------------------------------------------------------------------------------------------------------
tone3                ' GET ATTENUATION VALUE             ' same as previous
                     mov    tab_ptr, #att_table          
                     add    tab_ptr, psg_att_3           
                     movs   :read3, tab_ptr              
:read3               mov    att_tmp_3, 0                 
                    ' SWITCH OSCILLATOR                  
                     cmp    psg_period_3, #1          wc        
                if_c mov    psg_out_3, att_tmp_3         
                if_c jmp    #noise                       
                    ' CALCULATE OSCILLATOR               
                     sub    psg_cnt_3, psg_SubUnit    wc                                          
                if_c mov    psg_cnt_3, psg_period_3      
                if_c xor    psg_osc_3, #1             wz 
                if_c negz   psg_out_3, att_tmp_3                                                                                                                                      
' -------------------------------------------------------------------------------------------------------------
noise          ' GET ATTENUATION VALUE               
                     mov    tab_ptr, #att_table          ' create a pointer to the attenuation table
                     add    tab_ptr, psg_att_4           ' add the correct index to that pointer
                     movs   :readN, tab_ptr              ' move the pointer to the source field of instruction at :read position
:readN               mov    att_tmp_4, 0                 ' moved the value in the source field ( 0 is a place holder ) to tab_data
               ' REFRESH COUNTER AND GET NOISE TYPE   
                     sub    psg_cnt_4, psg_SubUnit    wc
               if_nc jmp    #mixer
                     mov    psg_cnt_4, psg_period_4 
                     testn  TapBits, #1               wz ' check noise type ( $1 & $1 = 1, $9 & 1 = $8 ) 
                if_z jmp    #periodicNoise               ' if 0 is a periodic noise 
               ' COMPUTE WHITE NOISE    
whiteNoise           test   ShiftReg, TapBits         wc ' test TapBits and shift reg, c = 1 if odd, c = 0 if even
                     mov    shift_tmp, #0
                     muxc   shift_tmp, #1                ' set lowest bit to c state
                     jmp    #updateShiftReg              ' update shift register   
               ' COMPUTE PERIODIC NOISE            
periodicNoise        mov    shift_tmp, ShiftReg          ' just take the lowes bit into a temp value
                     and    shift_tmp, #1                
               ' REFRESH SHIFT REGISTER AND COMPUTE OUTPUT                              
updateShiftReg       shl    shift_tmp, noiseMSB          ' shift result bit to the highest bit
                     shr    ShiftReg, #1                 ' shift register one place to right
                     or     ShiftReg, shift_tmp          ' merge temp bit and actual shift register
                     mov    shift_tmp, ShiftReg
                     and    shift_tmp, #1             wz ' check the lowest bit, if 0, z=1, if 1 z=0
                if_z mov    psg_out_4, #0
               if_nz mov    psg_out_4, att_tmp_4         ' use attenuation according to the bit               

' MIX VALUES TO OUTPUT MASTER CHANNEL ---------------------------------------------------------------
mixer                                                                      
                     mov    out_audio, dc_offset         ' add DC offset ( PWM 50% duty cycle )      
                     add    out_audio, psg_out_1         ' add out channel 1 ( Tone 1 )
                     add    out_audio, psg_out_2         ' add out channel 2 ( Tone 2 )
                     add    out_audio, psg_out_3         ' add out channel 3 ( Tone 3 )
                     add    out_audio, psg_out_4         ' add out channel 4 ( Noise )
                                                                                             
output_audio         mov    frqa, out_audio              ' update master output frequence to change the duty cycle

                     wrlong ShiftReg, chip_system        ' update LFSR into the shared memory
                                           
                    waitcnt cnt_accu, cnt_for_sample     ' wait for a "sample" period
                                                                    
                     jmp    #refresh                     ' continue rendering samples
                     
' VARIABLES -----------------------------------------------------------------------------------------
Audio_pin           long    0
chip_system         long    0                            ' pointer to cog parameters  
cnt_precision       long    1                                               
cnt_accu            long    0                            ' counter accumulator                         
cnt_for_sample      long    trunc( 358.37 * 2.0 )        ' cycles to wait each render cylce *** Probably to FIX ***           

{                                                                                                                                         
 This value is obtained followinf the "hardware approach"
 PSG:
 3579545 ( main PAL clock. ) / 16 ( PSG clk divider ) / 109 Hz ( Minimum frequency of PSG ) = 2052.49
 this is ful period, but for changing polarity of a wave we need the half of it
 2052.49 / 2 = 1026.45, usually rounded to 1024 ( $3FF )
 So we have an exactly correspondance to 1 unit to decrement to 1 clock cycle
 In propeller we need to calculate how many cycles it need to run to obtain a 109 Hz wave
 with a 1024 period counter.

 PROPELLER:
 80000000 ( cog freq. ) / 109 Hz ( Minimum frequency of PSG ) = 733955

 We now have how many cycles there are in the audio wave of 109 Hz
 and dividing by 2

 733955 / 2 = 366977,5 
 these are the cycles that the cog must run in a period of $3FF
 so if we need the number of cycles for each unit, we need just to divide
 366977 / $3FF = 358,37
 Here we are, we should wait for this number of cycles each period decrement

 I decided to double this period to have more resolution for calculation, in fact if at
 the end of the rendering cycle cog need to wait the counter has reached a cnt +358,37 cycle
 it means that it is supposed to run more or less ( 358,37 / 4 ) 89 instructions ( 4 cycle each
 not considering hub access )
 If you have more instructions it doesn't work.
}

psg_SubUnit         long    4           ' unit to subtract from each period unit *** Probably to FIX ***
                                                                 
t1                  long    0           ' temp variables
t2                  long    0
att_tmp_1           long    0
att_tmp_2           long    0
att_tmp_3           long    0
att_tmp_4           long    0 
per_tmp             long    0

shift_tmp           long    0           ' noise handlers
shift_out           long    0
psg_noise_Reg       long    0
TapBits             long    0
ShiftReg            long    0                                       
LSFRMask            long    $FFFF
noiseMSB            long    15

tab_ptr             long    0           ' pointer for the use of the table
att_table           long    $1fffffff   ' lookup tables to emulate logaritmic output volume
                    long    $19694450  
                    long    $142dd520  
                    long    $10063010  
                    long    $0cb98c10  
                    long    $0a1ad170  
                    long    $08063130  
                    long    $065f3bf0  
                    long    $050f5ce0  
                    long    $0404a5d0  
                    long    $0330d960  
                    long    $0288a8f0  
                    long    $020319d0  
                    long    $01990aa0  
                    long    $0144d1e0  
                    long    $00000000

psg_cnt_1           long    0          ' counter channel accumulators
psg_cnt_2           long    0            
psg_cnt_3           long    0
psg_cnt_4           long    0
                                                                                 
psg_osc_1           long    1          ' channel oscillators sign                     
psg_osc_2           long    1                 
psg_osc_3           long    1
psg_osc_4           long    1 
                                         
psg_att_1           long    $f         ' channel attenuators                      
psg_att_2           long    $f               
psg_att_3           long    $f
psg_att_4           long    $f                                                    

psg_period_1        long    $0         ' channel periods       
psg_period_2        long    $0    
psg_period_3        long    $0
psg_period_4        long    $0
Period_mask         long    $3FF                                    
 
psg_out_1           long    0          ' output values      
psg_out_2           long    0    
psg_out_3           long    0     
psg_out_4           long    0
out_audio           long    0          ' master output value
dc_offset           long    $8000_0000 ' PWM duty cycle of 50%


{{                   
 ______________________________________________________________________________________________________________________________
|                                                   TERMS OF USE: MIT License                                                  |                                                            
|______________________________________________________________________________________________________________________________|
|Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    |     
|files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    |
|modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software|
|is furnished to do so, subject to the following conditions:                                                                   |
|                                                                                                                              |
|The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.|
|                                                                                                                              |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          |
|WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         |
|COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   |
|ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         |
 ------------------------------------------------------------------------------------------------------------------------------ 
}}